#include "tree2files.h"

static int version_flag = 0;

static struct option long_options[] = {
  {"version",     no_argument,       &version_flag, 1},
  {"help",        no_argument,       NULL, 'h'},
  {NULL, 0, NULL, 0}
};

static void strnchomp(char *dest, size_t n) {
  size_t len;
#ifdef HAVE_STRNLEN
  len = strnlen(dest, n)-1;
#else
  len = strlen(dest)-1;
#endif
  if (len > 0 && dest[len] == '\n') {
    dest[len]='\0';
  }
}

static void push(struct node ** tos, struct node **bos, char * name) {

  /* new node */
  struct node * dir;
  size_t        len;

  dir = malloc(sizeof(struct node));
  if (!dir) {
    fprintf(stderr, "Error: out of memory\n");
    exit(EXIT_FAILURE);
  }
#ifdef HAVE_STRNLEN
  len = strnlen(name, MAX_LINE_LENGTH)+1;
#else
  len = strlen(name)+1;
#endif
  dir->name = malloc(len);
  if (!dir->name) {
    fprintf(stderr, "Error: out of memory\n");
    exit(EXIT_FAILURE);
  }    
  strncpy(dir->name, name, len);
  dir->next = NULL;
  dir->prev = NULL;

  /* attach */
  dir->next  = *tos;
  if (*tos) {
    (*tos)->prev = dir;
  } else {
    *bos = dir;
  }
  *tos = dir;

}

static void pop(struct node ** tos) {
  struct node * tmp = *tos;
  if ((*tos)->next != NULL) {
    (*tos)->next->prev = NULL;
  }
  *tos = (*tos)->next;
  if (tmp &&
      tmp->name) {
    free(tmp->name);
  }
  free(tmp);
}
   
static void print_rev_stack(struct node *bos) {
  struct node * cur = bos;
  if (cur) {
    printf("%s", cur->name);
    cur = cur->prev;
    while (cur) {
      printf("%s", cur->name);
      cur = cur->prev;
    }
  }
}

static void read_tree() {

  char         * line       = NULL;
  int            bytes_read = 0;
  size_t         length     = 0;
  size_t         indent     = 0;
  size_t         old_indent = 0;
  size_t         len        = 0;

  struct node  * tos        = NULL;
  struct node  * bos        = NULL;

  /* read line */
  bytes_read = getline(&line, &length, stdin);
  
  while (bytes_read > 0 &&
         line       != NULL) {

    /* process the line */

    strnchomp(line, MAX_LINE_LENGTH);
    old_indent = indent;
    indent     = 0;
    while (indent       < MAX_LINE_LENGTH &&
	   indent       < length          &&
	   line[indent] == ' ') {
      indent++;
    }

    while (old_indent > indent) {
      pop(&tos);
      old_indent--;
    }
    
#ifdef HAVE_STRNDUP
    len = strnlen(line, MAX_LINE_LENGTH);
#else
    len = strlen(line);
#endif

    print_rev_stack(bos);

    printf("%s\n", line+indent);

    if (line[len-1] == '/') {
      /* directory -> push */
      push(&tos, &bos, line+indent);
      indent++;
    }

    /* read line */
    bytes_read = getline(&line, &length, stdin);

  }

  /* deallocate the stack */
  while (tos) {
    pop(&tos);
  }
  
  if (line) {
    free(line);
  }

}

static void usage() {
  
  printf("\nUsage: [OPTIONS] [AUTHFILE]\n\n");
  printf("Options\n");
  printf(" -h, --help         this help message\n");
  printf(" --version          prints the version of svncheck\n");
  printf("\n\nPlease see the tree2files(1) man page for full documentation\n");

}

int main(int argc, char **argv) {
  
  int c;

  while (1) {

    /* getopt_long stores the option index here. */
    int option_index = 0;
     
    c = getopt_long (argc, argv, "h", long_options, &option_index);
     
    /* Detect the end of the options. */
    if (c == -1)
      break;
     
    switch (c) {

    case 'h':
      usage();
      exit(EXIT_SUCCESS);

    case '?':
      usage();
      /* getopt_long already printed an error message. */
      exit(EXIT_FAILURE);

    case 0:
      break;

    default:
      abort ();
    }

  }

  if (version_flag) {
    printf("tree2files %s\n", PACKAGE_VERSION);
    exit(EXIT_SUCCESS);
  }

  read_tree();

  return 0;

}
